package skyrix;

import java.io.IOException;
import java.net.URL;
import java.util.Collection;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Vector;

import org.apache.xmlrpc.XmlRpcClient;
import org.apache.xmlrpc.XmlRpcException;

/**
 * @author <a href="mailto:loutchan@skyrix.com">Eileen Loutchan</a>
 * @version $Revision: $
 */
public class AppointmentManager
{

  /**Referenz auf den XML-RPC Client.*/
	private XmlRpcClient client = null;

  /**Konstruktor.
   * 
   * <p>Instanziert die Klasse.</p>
   */
	public AppointmentManager(XmlRpcClient client)
	{
		this.client = client;			
	}

	/**Methode gibt alle Termine zurck, die auf die bergebenden
	 * Kriterien passen.
	 * 
	 * @param startDate Anfangszeitraum.
	 * @param endDate Endzeitraum.
	 * @return Liste mit den Terminen.
	 */
	public Collection getAllAppointments(Date startDate, Date endDate) throws XmlRpcException, IOException
	{
    if(startDate == null) return null;
    
		LinkedList allAppointments = new LinkedList();
		Vector argv = new Vector();

		Hashtable hash = new Hashtable();
		Hashtable dateHash = new Hashtable();
		dateHash.put("startDate", startDate);
		if(endDate != null)
		dateHash.put("endDate", endDate);

		hash.put("qualifier", dateHash);

		argv.addElement(hash);

		Object ret = client.execute("appointment.fetch", argv);
		if (ret instanceof Vector)
		{
			Vector retVector = (Vector) ret;
			Enumeration en = retVector.elements();
			while (en.hasMoreElements())
			{
				Object data = en.nextElement();
				if (data instanceof Hashtable)
				{
					Appointment appointment = Appointment.createAppointment((Hashtable) data);
					if (appointment != null)
						allAppointments.add(appointment);
				}
			}
       
      System.out.println("getAllAppointments(..): "+allAppointments);
			return allAppointments;
		}

		return null;
	}

	/**Methode gibt alle Termin-IDs fr die Termine, auf die die bergebenden
	 * Kriterien passen, zurck.
	 * 
	 * <p>Die ID wird als String zurckgegeben.</p>
	 * 
	 * @param startDate Anfangszeitraum.
	 * @param endDate Endzeitraum.
	 * @return Liste mit den IDs.
	 */
	public Collection getAllAppointmentIDs(Date startDate, Date endDate) throws XmlRpcException, IOException
	{
    if(startDate == null) return null;

		LinkedList allAppointments = new LinkedList();
		Vector argv = new Vector();

		Hashtable hash = new Hashtable();
		Hashtable dateHash = new Hashtable();
		dateHash.put("startDate", startDate);
		if(endDate != null)
		  dateHash.put("endDate", endDate);

		hash.put("qualifier", dateHash);

		argv.addElement(hash);

		Object ret = client.execute("appointment.fetchIds", argv);

		if (ret instanceof Vector)
		{
			Vector retVector = (Vector) ret;
			Enumeration en = retVector.elements();
			while (en.hasMoreElements())
			{
				Object obj = en.nextElement();

				if (obj instanceof String)
				{
					String id = (String) obj;
					if (id != null)
						allAppointments.add(id);
				}
			}

      System.out.println("getAllAppointmentIDs(..): "+allAppointments);
			return allAppointments;
		}

		return null;
	}

	/**Methode gibt den Termin zu der bergebenden ID zurck.
	 * 
	 * @param id Identifiziert einen Termin.
	 * @return Termin zu der ID.
	 */
	public Appointment getAppointmentByID(String id) throws XmlRpcException, IOException
	{
		if(id == null) return null;
		
		Vector argv = new Vector();
		argv.addElement(id);

		Object ret = client.execute("appointment.getById", argv);

		if (ret instanceof Hashtable)
		{
      System.out.println("getAppointmentByID(..): "+ret);
			return Appointment.createAppointment((Hashtable) ret);
		}

		return null;
	}

	/**Methode gibt die Termine zu der bergebenden Liste mit den 
	 * IDs zurck.
	 * 
	 * <p>Methode funktioniert nicht,
	 * Server kann zur Zeit nur eine ID bearbeiten,
	 * benutze an dieser Stelle getAppointmentById(..).</p>
	 * 
	 * @param id Liste mit den IDs zu den Terminen.
	 * @return Liste mit Appointment Objekten.
	 */
	public Collection getAppointmentsByIDList(Collection idList) throws XmlRpcException, IOException
	{
		if (idList.isEmpty())	return null;

		LinkedList allAppointments = new LinkedList();
		Vector argv = new Vector(idList);

		Object ret = client.execute("appointment.getById", argv);
//System.out.println("Ret ist: "+ret.getClass());
//System.out.println("Hash ist: "+((Hashtable)ret));

		if (ret instanceof Vector)
		{
			Vector retVector = (Vector) ret;
			Enumeration en = retVector.elements();
			while (en.hasMoreElements())
			{
				Object data = en.nextElement();
				if (data instanceof Hashtable)
				{
					Appointment appointment = Appointment.createAppointment((Hashtable) data);
					if (appointment != null)
						allAppointments.add(appointment);
				}
			}
      
      System.out.println("getAppointmentsByIDList(..): "+allAppointments);
			return allAppointments;
		}

		return null;
	}

	/**Methode legt einen neuen Termin an.
	 * 
	 * @param startDate Anfangszeitpunkt.
	 * @param endDate Endzeitpunkt.
	 * @param name Name des Termins.
	 * @return Gibt den angelegten Termin zurck. 
	 */
	public Appointment insertAppointment(Date startDate, Date endDate, String title, String comment) throws XmlRpcException, IOException
	{
		if (title.trim().equals("") || startDate == null || endDate == null)
			return null;

		Vector argv = new Vector();
		Hashtable hash = new Hashtable();
		hash.put("startDate", startDate);
		hash.put("endDate", endDate);
		hash.put("title", title);
		if(comment != null)
		hash.put("comment", comment);
		argv.addElement(hash);

		Object ret = client.execute("appointment.insert", argv);
		if (ret instanceof Hashtable)
		{
      System.out.println("insertAppointment(..): "+ret);
			return Appointment.createAppointment((Hashtable) ret);
		}

		return null;
	}

	/**Methode aktualisiert das bergebende Termin-Objekt.
	 * 
	 * <p>Das Id Attribut muss zwingend gesetzt sein.</p>
	 * 
	 * @param appointment Termin-Objekt
	 */
	public Appointment updateAppointment(Appointment appointment) throws XmlRpcException, IOException
	{
		if (appointment.getID() == null)
			return null;

		Vector argv = new Vector();
    
		Hashtable hash = new Hashtable();
		hash.put("id", appointment.getID());
		if (appointment.getStartDate() != null)
			hash.put("startDate", appointment.getStartDate());
		if (appointment.getEndDate() != null)
			hash.put("endDate", appointment.getEndDate());
		if (appointment.getTitle() != null)
			hash.put("title", appointment.getTitle());

		argv.addElement(hash);

		Object ret = client.execute("appointment.update", argv);

		if (ret instanceof Hashtable)
		{
      System.out.println("updateAppointment(..): "+ret);
			return Appointment.createAppointment((Hashtable) ret);
		}

		return null;
	}

	/**Methode lscht das bergebende Termin-Objekt.
	 * 
	 * @param appointment Termin-Objekt
	 */
	public void deleteAppointment(Appointment appointment) throws XmlRpcException, IOException
	{
		if (appointment.getID() == null)
			return;

		Vector argv = new Vector();
		Hashtable hash = new Hashtable();
		hash.put("id", appointment.getID());

		argv.addElement(hash);

		Object ret = client.execute("appointment.delete", argv);
	}

	/**Methode setzt die Teilnehmer zu einem Termin.
	 * 
	 * @param appointment Termin dem Teilnehmer zugeordnet werden sollen.
	 * @param participants Liste mit Personen - Objekten.
	 * @return Gibt einen Termin mit den gesetzten Teilnehmern zurck.
	 */
	public Appointment setParticipans(Appointment appointment, Collection participants) throws XmlRpcException, IOException
	{
    if (appointment.getID() == null || participants.isEmpty())
      return null;

		Vector loginv = new Vector();
		loginv.addElement("eileen");

		Object loginID = null;
		try
		{
			loginID = client.execute("account.getByLogin", loginv);
		}
		catch (IOException ex)
		{
			ex.printStackTrace();
		}
		catch (XmlRpcException ex)
		{
			ex.printStackTrace();
		}

    if(loginID != null)
    {
			Vector argv = new Vector();
	    Hashtable hash = new Hashtable();
	    hash.put("id", appointment.getID());      
	    argv.addElement(hash);
	    
	    Vector personID = new Vector();   
	    Iterator personIterator = participants.iterator();
	    while(personIterator.hasNext())
	    {
	    	Object obj = personIterator.next();
	    	if(obj instanceof Person)
	    	{
	    		Person person = (Person)obj;
			    personID.addElement(person.getID());
	    	}
	    }
			personID.addElement(loginID); // eigener Login
			
	    argv.addElement(personID);
	
			Object ret = client.execute("appointment.setParticipants", argv);

			if(ret instanceof Hashtable)
			{
				System.out.println("setParticipans(..): "+ret);
				//Ergebnis enthlt nicht die angelegten Teilnehmer
				//zu den IDs mssen die Personen Objekte erstellt werden
				//die Liste mit den IDs muss im Appointment Objekt ausgelesen werden 
				return Appointment.createAppointment((Hashtable) ret);
			}
    }
    
		return null;
	}

  /**Methode gibt alle Personen zurck.
   * 
   * <p>Die Personen knnen als Teilnehmer fr einen
   * Termin gesetzt werden.</p>
   * @return Collection Liste mit Person - Objekten
   */
  public Collection getAllPersons() throws XmlRpcException, IOException
  {
  	LinkedList allPersons = new LinkedList();
		Vector argv = new Vector();
		argv.addElement("name like '*'");

		Object ret = client.execute("person.fetch", argv);

		if(ret instanceof Vector)
		{
			Vector retVector = (Vector) ret;
			Enumeration en = retVector.elements();
			while (en.hasMoreElements())
			{
				Object data = en.nextElement();
				if (data instanceof Hashtable)
				{
					Person person = Person.createPerson((Hashtable) data);
					if (person != null)
					  allPersons.add(person);
				}
			}
       
			System.out.println("getAllPersons(..): "+allPersons);
			return allPersons;
		}
		
		return null;  	
  }

  /**Nur fr die Entwicklung.*/
	public static void main(String[] args)
	{
		try
		{
			XmlRpcClient client = new XmlRpcClient(new URL("http://marvin:30147/RPC2"));
			client.setBasicAuthentication("root", "root");

			AppointmentManager app = new AppointmentManager(client);
			//Appointment o = app.insertAppointment(new Date(103, 3, 11), new Date(103, 4, 11), "Hallle", "Inhalt");
			//app.getAllAppointmentIDs(new Date(103, 0, 26), new Date(103, 0, 30));
			//app.getAllAppointments(new Date(103, 0, 26), null);
			//app.getAppointmentByID(o.getID());
		
			//app.getAppointmentsByIDList(app.getAllAppointmentIDs(new Date(103, 0, 22), new Date(103, 0, 29)));
		  
		  //o.setTitle("Update");
		  //app.updateAppointment(o);
		  
			//app.deleteAppointment(o);
		  //Collection persons = app.getAllPersons();
		  //app.setParticipans(o, persons);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
}
